# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import pickle
import numpy as np
from h5py import string_dtype
from numbers import Number

import numpy.typing as npt
from typing import Any
from typing import Mapping
from typing import TYPE_CHECKING

from ch5mpy.objects import File
from ch5mpy.objects import Group
from ch5mpy.utils import is_sequence

if TYPE_CHECKING:
    from ch5mpy import H5Array


# ====================================================
# code
def write_attribute(group: Group, name: str, obj: Any) -> None:
    """Write a simple object as a H5 group attribute."""
    try:
        group.attrs[name] = "None" if obj is None else obj

    except TypeError:
        # since <obj> cannot be stored directly, we pickle it
        # here we use numpy's void type to allow storing bytes generated by pickle
        group.attrs[name] = np.void(pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL))


def write_attributes(group: Group, **kwargs: Any) -> None:
    """Write multiple object as h5 group attributes."""
    for name, obj in kwargs.items():
        write_attribute(group, name, obj)


def _store_dataset(
    loc: Group | File, name: str, array: npt.NDArray[Any] | H5Array[Any]
) -> None:
    """Store a dataset."""
    if np.issubdtype(array.dtype, np.str_):
        loc.create_dataset(name, data=array.astype(object), dtype=string_dtype())

    else:
        loc.create_dataset(name, data=array)


def write_dataset(loc: Group | File, name: str, obj: Any) -> None:
    """Write an array-like object to a H5 dataset."""
    if isinstance(obj, Mapping):
        group = loc.create_group(name)
        write_datasets(group, **obj)

    else:
        # cast to np.array if needed (to get shape and dtype)
        array = np.array(obj) if not hasattr(obj, "shape") else obj

        if name in loc.keys():
            if loc[name] is array:
                # this exact dataset is already stored, do nothing
                return

            elif loc[name].shape == array.shape and loc[name].dtype == array.dtype:
                # a similar array already exists, simply copy the data
                loc[name][()] = array

            else:
                # a different array was stored, delete it before storing the new array
                del loc[name]
                _store_dataset(loc, name, array)

        else:
            _store_dataset(loc, name, array)


def write_datasets(loc: Group | File, **kwargs: Any) -> None:
    """Write multiple array-like objects to H5 datasets."""
    for name, obj in kwargs.items():
        write_dataset(loc, name, obj)


def write_object(loc: Group | File, name: str, obj: Any) -> None:
    """Write any object to a H5 file."""
    if isinstance(obj, Mapping):
        group = loc.create_group(name)
        write_objects(group, **obj)

    elif is_sequence(obj) or isinstance(obj, (Number, str)):
        write_dataset(loc, name, obj)

    else:
        raise NotImplementedError


def write_objects(loc: Group | File, **kwargs: Any) -> None:
    """Write multiple objects of any type to a H5 file."""
    for name, obj in kwargs.items():
        write_object(loc, name, obj)



# coding: utf-8

# ====================================================
# imports
import pytest
import numpy as np
from pathlib import Path

from ch5mpy import File
from ch5mpy import H5Dict
from ch5mpy import H5Mode
from ch5mpy import H5Array
from ch5mpy import write_object


# ====================================================
# code
@pytest.fixture
def backed_dict() -> H5Dict:
    data = {
        "a": 1,
        "b": [1, 2, 3],
        "c": {"d": "test", "e": np.arange(100)},
        "f": np.zeros((10, 10, 10)),
    }

    with File("backed_dict", H5Mode.WRITE_TRUNCATE) as h5_file:
        write_object(h5_file, "uns", data)

    yield H5Dict(File("backed_dict", H5Mode.READ_WRITE)["uns"])

    Path("backed_dict").unlink()


def test_backed_dict_creation(backed_dict):
    assert isinstance(backed_dict, H5Dict)


def test_backed_dict_can_iterate_through_keys(backed_dict):
    assert list(iter(backed_dict)) == ["a", "b", "c", "f"]


def test_backed_dict_has_correct_keys(backed_dict):
    assert list(backed_dict.keys()) == ["a", "b", "c", "f"]


def test_backed_dict_can_get_regular_values(backed_dict):
    assert backed_dict["a"] == 1


def test_backed_dict_can_get_string(backed_dict):
    assert backed_dict["c"]["d"] == "test"


def test_backed_dict_can_get_with_getattr(backed_dict):
    assert isinstance(backed_dict.c.e, H5Array)


def test_backed_dict_should_return_string(backed_dict):
    assert isinstance(backed_dict["c"]["d"], str)


def test_backed_dict_gets_nested_backed_dicts(backed_dict):
    assert isinstance(backed_dict["c"], H5Dict)


def test_backed_dict_has_correct_values(backed_dict):
    values_list = list(backed_dict.values())

    assert (
        values_list[0] == 1
        and np.array_equal(values_list[1], [1, 2, 3])
        and list(values_list[2].keys()) == ["d", "e"]
    )


def test_backed_dict_has_correct_items(backed_dict):
    assert list(backed_dict.items())[0] == ("a", 1)


def test_backed_dict_can_set_regular_value(backed_dict):
    backed_dict["a"] = 5

    assert np.all(backed_dict["a"] == 5)


def test_backed_dict_can_set_array_value(backed_dict):
    backed_dict["b"][1] = 6

    assert np.all(backed_dict["b"] == [1, 6, 3])


def test_backed_dict_can_set_new_regular_value(backed_dict):
    backed_dict["x"] = 9

    assert backed_dict["x"] == 9


def test_backed_dict_can_set_new_array(backed_dict):
    backed_dict["y"] = np.array([1, 2, 3])

    assert np.all(backed_dict["y"] == [1, 2, 3])


def test_backed_dict_can_set_new_dict(backed_dict):
    backed_dict["z"] = {"l": 10, "m": [10, 11, 12], "n": {"o": 13}}

    assert (
        isinstance(backed_dict["z"], H5Dict)
        and backed_dict["z"]["l"] == 10
        and np.all(backed_dict["z"]["m"] == [10, 11, 12])
        and isinstance(backed_dict["z"]["n"], H5Dict)
        and backed_dict["z"]["n"]["o"] == 13
    )


def test_backed_dict_can_delete_regular_value(backed_dict):
    del backed_dict["a"]

    assert "a" not in backed_dict.keys()


def test_backed_dict_can_delete_array(backed_dict):
    del backed_dict["b"]

    assert "b" not in backed_dict.keys()


def test_backed_dict_can_delete_dict(backed_dict):
    del backed_dict["c"]

    assert "c" not in backed_dict.keys()


def test_backed_dict_can_close_file(backed_dict):
    backed_dict.close()

    with pytest.raises(ValueError):
        _ = backed_dict["x"]


def test_backed_dict_copy_should_be_regular_dict(backed_dict):
    c = backed_dict.copy()

    assert isinstance(c, dict)


def test_backed_dict_copy_should_have_same_keys(backed_dict):
    c = backed_dict.copy()

    assert c.keys() == backed_dict.keys()


def test_backed_dict_copy_nested_backed_dict_should_be_dict(backed_dict):
    c = backed_dict.copy()

    assert isinstance(c["c"], dict)


def test_backed_dict_copy_dataset_proxy_should_be_array(backed_dict):
    c = backed_dict.copy()

    assert isinstance(c["b"], H5Array)



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import numpy as np
from numbers import Number

import numpy.typing as npt
from typing import Any
from typing import Iterable
from typing import TYPE_CHECKING

import ch5mpy
from ch5mpy.h5array.inplace import iter_chunks_2
from ch5mpy.h5array.functions.apply import apply_2
from ch5mpy.h5array.functions.implement import implements


if TYPE_CHECKING:
    from ch5mpy import H5Array


# ====================================================
# code
def _cast_H5Array(obj: Any) -> H5Array[Any]:
    return obj                                                                             # type: ignore[no-any-return]


def ensure_h5array_first(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                         x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any]) \
        -> tuple[H5Array[Any], npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any]]:
    if not isinstance(x1, ch5mpy.H5Array):
        return _cast_H5Array(x2), x1

    return _cast_H5Array(x1), x2


@implements(np.array_equal)
def array_equal(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                equal_nan: bool = False) -> bool:
    x1, x2 = ensure_h5array_first(x1, x2)

    # case 0D
    if isinstance(x2, Number):
        if x1.ndim:
            return False

        return np.array_equal(x1, x2, equal_nan=equal_nan)                                  # type: ignore[arg-type]

    # case nD
    if not isinstance(x2, (np.ndarray, ch5mpy.H5Array)):
        x2 = np.array(x2)

    if x1.shape != x2.shape:
        return False

    for index, chunk_x1, chunk_x2 in iter_chunks_2(x1, x2):
        if not np.array_equal(chunk_x1, chunk_x2):                                              # type: ignore[arg-type]
            return False

    return True


@implements(np.greater)
def greater(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
            x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
            out: H5Array[Any] | npt.NDArray[Any] | None = None,
            where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
            dtype: npt.DTypeLike | None = None) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(np.greater, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=False)


@implements(np.greater_equal)
def greater_equal(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                  x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                  out: H5Array[Any] | npt.NDArray[Any] | None = None,
                  where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
                  dtype: npt.DTypeLike | None = None) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(np.greater_equal, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=False)


@implements(np.less)
def less(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
         x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
         out: H5Array[Any] | npt.NDArray[Any] | None = None,
         where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
         dtype: npt.DTypeLike | None = None) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(np.less, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=False)


@implements(np.less_equal)
def less_equal(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
               x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
               out: H5Array[Any] | npt.NDArray[Any] | None = None,
               where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
               dtype: npt.DTypeLike | None = None) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(np.less_equal, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=False)


@implements(np.equal)
def equal(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
          x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
          out: H5Array[Any] | npt.NDArray[Any] | None = None,
          where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
          dtype: npt.DTypeLike | None = None) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(np.equal, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=False)


@implements(np.not_equal)
def not_equal(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
              x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
              out: H5Array[Any] | npt.NDArray[Any] | None = None,
              where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
              dtype: npt.DTypeLike | None = None) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(np.not_equal, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=False)


@implements(np.add)
def add(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
        x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
        out: H5Array[Any] | npt.NDArray[Any] | None = None,
        where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
        dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.add, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.multiply)
def multiply(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
             x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
             out: H5Array[Any] | npt.NDArray[Any] | None = None,
             where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
             dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.multiply, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.divide)
def divide(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
           x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
           out: H5Array[Any] | npt.NDArray[Any] | None = None,
           where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
           dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.divide, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.power)
def power(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
          x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
          out: H5Array[Any] | npt.NDArray[Any] | None = None,
          where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
          dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.power, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.subtract)
def subtract(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
             x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
             out: H5Array[Any] | npt.NDArray[Any] | None = None,
             where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
             dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.subtract, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.true_divide)
def true_divide(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                out: H5Array[Any] | npt.NDArray[Any] | None = None,
                where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
                dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.true_divide, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.floor_divide)
def floor_divide(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                 x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                 out: H5Array[Any] | npt.NDArray[Any] | None = None,
                 where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
                 dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.floor_divide, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.float_power)
def float_power(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                out: H5Array[Any] | npt.NDArray[Any] | None = None,
                where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
                dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.float_power, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.fmod)
def fmod(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
         x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
         out: H5Array[Any] | npt.NDArray[Any] | None = None,
         where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
         dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.fmod, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.mod)
def mod(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
        x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
        out: H5Array[Any] | npt.NDArray[Any] | None = None,
        where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
        dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.mod, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.maximum)
def maximum(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
            x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
            out: H5Array[Any] | npt.NDArray[Any] | None = None,
            where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
            dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.maximum, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.fmax)
def fmax(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
         x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
         out: H5Array[Any] | npt.NDArray[Any] | None = None,
         where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
         dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.fmax, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.minimum)
def minimum(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
            x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
            out: H5Array[Any] | npt.NDArray[Any] | None = None,
            where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
            dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.minimum, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.fmin)
def fmin(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
         x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
         out: H5Array[Any] | npt.NDArray[Any] | None = None,
         where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
         dtype: npt.DTypeLike | None = None) -> Any:
    return apply_2(np.fmin, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.logical_and)
def logical_and(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                out: H5Array[Any] | npt.NDArray[Any] | None = None,
                where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
                dtype: npt.DTypeLike | None = None) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(np.logical_and, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.logical_or)
def logical_or(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
               x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
               out: H5Array[Any] | npt.NDArray[Any] | None = None,
               where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
               dtype: npt.DTypeLike | None = None) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(np.logical_or, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.logical_not)
def logical_not(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                out: H5Array[Any] | npt.NDArray[Any] | None = None,
                where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
                dtype: npt.DTypeLike | None = None) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(np.logical_not, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)


@implements(np.logical_xor)
def logical_xor(x1: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                x2: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
                out: H5Array[Any] | npt.NDArray[Any] | None = None,
                where: npt.NDArray[np.bool_] | Iterable[bool] | bool = True,
                dtype: npt.DTypeLike | None = None) -> Any:
    if dtype is None:
        dtype = bool

    return apply_2(np.logical_xor, *ensure_h5array_first(x1, x2), out=out, dtype=dtype, where=where, default=x1)



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import numpy as np
from numbers import Number
from collections.abc import KeysView
from collections.abc import Iterable
from collections.abc import Collection
from collections.abc import MutableMapping
from h5py._hl.base import ItemsViewHDF5

from typing import Any
from typing import cast
from typing import Union
from typing import Iterator

import ch5mpy as hu
from ch5mpy.objects import File
from ch5mpy.objects import Group
from ch5mpy.objects import Dataset
from ch5mpy.write import write_object

# ====================================================
# code
H5DICT_CONTENT = Union["H5Dict", "hu.H5Array[Any]", Number, str]


def _get_in_memory(value: Any) -> Any:
    if isinstance(value, H5Dict):
        return value.copy()

    elif isinstance(value, Dataset):
        return value[()]

    return value


def _get_repr(items: ItemsViewHDF5[str, Group | Dataset[Any]]) -> str:
    return ", ".join([str(k) + ": " + (repr(_parse_value(v)) if not isinstance(v, Group) else "{...}")
                      for k, v in items])


class H5Dict(MutableMapping[str, 'H5DICT_CONTENT']):
    """Class for managing dictionaries backed on h5 files."""

    # region magic methods
    def __init__(self, file: File | Group):
        assert isinstance(file, (File, Group))

        self._file = file

    def __dir__(self) -> Iterable[str]:
        return dir(H5Dict) + list(self.keys())

    def __repr__(self) -> str:
        if self.is_closed:
            return "Closed H5Dict{}"

        return f"H5Dict{'{' + _get_repr(self._file.items()) + '}'}"

    def __setitem__(self, key: str, value: Any) -> None:
        write_object(self._file, key, value)

    def __delitem__(self, key: str) -> None:
        del self._file[key]

    def __getitem__(self, key: str) -> H5DICT_CONTENT:
        return _parse_value(self._file[key])

    def __getattr__(self, item: str) -> H5DICT_CONTENT:
        return self.__getitem__(item)

    def __len__(self) -> int:
        return len(self._file.keys())

    def __iter__(self) -> Iterator[str]:
        return iter(self.keys())

    # endregion

    # region predicates
    @property
    def is_closed(self) -> bool:
        return not self._file.id.valid

    # endregion

    # region methods
    def keys(self) -> KeysView[str]:
        return self._file.keys()

    def values(self) -> H5DictValuesView:                           # type: ignore[override]
        # from typing import reveal_type
        return H5DictValuesView(self._file.values())

    def items(self) -> H5DictItemsView:                             # type: ignore[override]
        return H5DictItemsView(self._file.keys(), self._file.values())

    def close(self) -> None:
        self._file.file.close()

    def copy(self) -> dict[str, Any]:
        """
        Build an in-memory copy of this H5Dict object.
        """
        return {k: _get_in_memory(v) for k, v in self.items()}

    # endregion


def _parse_value(obj: Group | Dataset[Any]) -> H5DICT_CONTENT:
    """Parse a h5 object into a higher abstraction-level object."""
    if isinstance(obj, Group):
        # return Group as H5Dict
        return H5Dict(obj)

    elif isinstance(obj, Dataset):
        if obj.shape == ():
            # single value in dataset
            if np.issubdtype(obj.dtype, np.number) or np.issubdtype(
                obj.dtype, np.bool_
            ):
                # return numeric value
                return cast(Number, obj[()])

            else:
                # return string value
                return cast(str, obj.asstr()[()])

        # return dataset as H5Array
        return hu.H5Array(obj)

    else:
        raise ValueError(
            f"Got unexpected object of type '{type(obj)}' for key '{obj}'."
        )


class H5DictValuesView(Iterable[H5DICT_CONTENT]):
    """Class for iterating over values in an H5Dict."""

    # region magic methods
    def __init__(self, values: Collection[Group | Dataset[Any]]):
        self._values = values

    def __repr__(self) -> str:
        return f"{type(self).__name__}([{len(self._values)} values])"

    def __iter__(self) -> Iterator[H5DICT_CONTENT]:
        return map(_parse_value, self._values)

    # endregion


class H5DictItemsView(Iterable[tuple[str, H5DICT_CONTENT]]):
    """Class for iterating over items in an H5Dict."""

    # region magic methods
    def __init__(self, keys: Collection[str], values: Collection[Group | Dataset[Any]]):
        self._keys = keys
        self._values = values

    def __repr__(self) -> str:
        return f"{type(self).__name__}([{len(self._keys)} items])"

    def __iter__(self) -> Iterator[tuple[str, H5DICT_CONTENT]]:
        return zip(self._keys, map(_parse_value, self._values))

    # endregion



# coding: utf-8

# ====================================================
# imports
import numpy as np


# ====================================================
# code
def test_array_equal(small_array):
    assert np.array_equal(small_array, [1, 2, 3, 4, 5])


def test_sum(array):
    assert np.sum(array) == 4950


def test_sum_with_initial_value(array):
    assert np.sum(array, initial=1) == 4951


def test_sum_with_output(array):
    out = np.array(0)
    assert np.sum(array, out=out) == 4950


def test_sum_with_initialized_output(array):
    out = np.array(10)
    assert np.sum(array, out=out) == 4950


def test_sum_with_output_and_diff_dtype(array):
    out = np.array(0, dtype=float)
    array += 0.5
    assert np.sum(array, out=out, dtype=int) == 4950


def test_sum_keepdims(array):
    assert np.array_equal(np.sum(array, keepdims=True), [[4950]])


def test_sum_with_axis(array):
    assert np.array_equal(np.sum(array, axis=0), [450, 460, 470, 480, 490, 500, 510, 520, 530, 540])


def test_sum_axis_0(small_large_array):
    small_large_array.MAX_MEM_USAGE = str(3 * small_large_array.dtype.itemsize)
    assert np.array_equal(np.sum(small_large_array, axis=0), np.array([[60, 63, 66, 69, 72],
                                                                       [75, 78, 81, 84, 87],
                                                                       [90, 93, 96, 99, 102],
                                                                       [105, 108, 111, 114, 117]]))


def test_sum_axis_2(small_large_array):
    small_large_array.MAX_MEM_USAGE = str(3 * small_large_array.dtype.itemsize)
    assert np.array_equal(np.sum(small_large_array, axis=2), np.array([[10., 35., 60., 85.],
                                                                       [110., 135., 160., 185.],
                                                                       [210., 235., 260., 285.]]))


def test_sum_where(small_large_array):
    assert np.array_equal(np.sum(small_large_array, axis=1, where=[True, False, False, True, True]),
                          np.array([[30., 0., 0., 42., 46.],
                                    [110., 0., 0., 122., 126.],
                                    [190., 0., 0., 202., 206.]]))


def test_sum_where_multiple_rows(small_large_array):
    small_large_array.MAX_MEM_USAGE = str(40 * small_large_array.dtype.itemsize)
    assert np.array_equal(np.sum(small_large_array, axis=1, where=[True, False, False, True, True]),
                          np.array([[30., 0., 0., 42., 46.],
                                    [110., 0., 0., 122., 126.],
                                    [190., 0., 0., 202., 206.]]))


def test_sum_where_few_elements(small_large_array):
    small_large_array.MAX_MEM_USAGE = str(3 * small_large_array.dtype.itemsize)
    assert np.array_equal(np.sum(small_large_array, axis=1, where=[True, False, False, True, True]),
                          np.array([[30., 0., 0., 42., 46.],
                                    [110., 0., 0., 122., 126.],
                                    [190., 0., 0., 202., 206.]]))


def test_all(array):
    assert not np.all(array)


def test_all_axis_keepdims(array):
    assert np.array_equal(np.all(array, axis=1, keepdims=True), np.array([[False], [True], [True], [True], [True],
                                                                          [True],  [True], [True], [True], [True]]))


def test_floor(array):
    array += 0.7
    assert np.array_equal(np.floor(array), np.arange(100.).reshape((10, 10)))


def test_ceil(array):
    array += 0.2
    assert np.array_equal(np.ceil(array), np.arange(1., 101.).reshape((10, 10)))


# def test_ceil_large(large_array):
#     np.ceil(large_array)


def test_trunc(array):
    array += 0.2
    assert np.array_equal(np.trunc(array), np.arange(100.).reshape((10, 10)))


def test_prod(array):
    assert np.array_equal(
        np.prod(array, axis=0, where=[True, False, False, False, False, True, True, True, True, True]),
        [0, 1, 1, 1, 1, 6393838623046875, 9585618768101376, 13865696119905399, 19511273389031424, 26853950884211451]
    )


def test_exp(array):
    assert np.array_equal(
        np.exp(array),
        np.exp(np.arange(100.).reshape((10, 10)))
    )


def test_exp_where(small_array):
    out = np.array([1., 1., 1., 1., 1.])
    np.exp(small_array, out, where=[True, False, False, True, True])

    assert np.array_equal(out, [np.exp(1), 1., 1., np.exp(4), np.exp(5)])


def test_expm1(small_array):
    assert np.array_equal(np.expm1(small_array), np.expm1([1, 2, 3, 4, 5]))


def test_isfinite(small_array):
    small_array[0] = np.inf
    assert np.array_equal(np.isfinite(small_array), [False, True, True, True, True])


def test_equal(small_array):
    assert np.all(np.equal(small_array, [1, 2, 3, 4, 5]))
    assert np.array_equal(np.equal(small_array, 2), [False, True, False, False, False])


def test_equal_where(small_array):
    assert np.array_equal(np.equal(small_array, [1, 2, 3, 3, 3], where=[True, True, False, False, True]),
                          [True, True, False, False, False])


def test_greater_element(small_array):
    assert np.array_equal(np.greater(small_array, 3), [False, False, False, True, True])


def test_multiply(small_array):
    assert np.array_equal(np.multiply(small_array, [2, 3, 4, 5, 6]), [2, 6, 12, 20, 30])


def test_multiply_where(small_array):
    assert np.array_equal(np.multiply(small_array, [2, 3, 4, 5, 6], where=[True, False, False, True, True]),
                          [2, 2, 3, 20, 30])


def test_isinf(small_array):
    small_array[1] = np.inf
    assert np.array_equal(np.isinf(small_array), [False, True, False, False, False])


def test_isinf_out(small_array):
    small_array[3] = np.inf
    out = np.array([1, 2, 3, 4, 5])
    np.isinf(small_array, where=[True, False, False, True, True], out=out)
    assert np.array_equal(out, [False, 2, 3, True, False])


def test_logical_and(small_array):
    assert np.array_equal(np.logical_and(small_array, [True, True, False, False, True]),
                          [True, True, False, False, True])



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

from numbers import Number
from typing import Generator
from typing import cast

import numpy as np

import numpy.typing as npt
from typing import Any
from typing import TypeVar
from typing import TYPE_CHECKING

import ch5mpy
from ch5mpy.h5array.slice import FullSlice

if TYPE_CHECKING:
    from ch5mpy import H5Array

# ====================================================
# code
_DT = TypeVar("_DT", bound=np.generic)
INF = np.iinfo(int).max


sizes = {"K": 1024, "M": 1024 * 1024, "G": 1024 * 1024 * 1024}


def get_size(s: int | str) -> int:
    value: int | None = None

    if isinstance(s, int):
        value = s

    elif s[-1] in sizes and s[:-1].lstrip("-").isdigit():
        value = int(s[:-1]) * sizes[s[-1]]

    elif s.isdigit():
        value = int(s)

    if value is None:
        raise ValueError(f"Unrecognized size '{s}'")

    if value <= 0:
        raise ValueError(f"Got invalid size ({value} <= 0).")

    return value


def get_chunks(max_memory_usage: int | str,
               shape: tuple[int, ...],
               itemsize: int) -> tuple[tuple[FullSlice, ...], ...]:
    # special case of 0D arrays
    if len(shape) == 0:
        raise ValueError("0D array")

    rev_shape = tuple(reversed(shape))
    nb_elements_chunk = int(get_size(max_memory_usage) / itemsize)

    # not enough max mem
    if nb_elements_chunk <= 1:
        raise ValueError(
            "Slicing is impossible because of insufficient allowed memory usage."
        )

    block_axes = int(np.argmax(~(np.cumprod(rev_shape + (np.inf,)) <= nb_elements_chunk)))
    size_block = (
        nb_elements_chunk // np.cumprod(rev_shape)[block_axes - 1]
        if block_axes
        else min(rev_shape[0], nb_elements_chunk)
    )

    if size_block == 0:
        block_axes = max(0, block_axes - 1)

    if block_axes == len(shape):
        # all array can be read at once
        return ((FullSlice.whole_axis(shape[0]),),)

    whole_axes = tuple(FullSlice.whole_axis(s) for s in rev_shape[:block_axes][::-1])
    iter_axis = rev_shape[block_axes]

    right_chunks = tuple(
        (FullSlice(s, min(s + size_block, iter_axis), 1, iter_axis), *whole_axes)
        for s in range(0, iter_axis, size_block)
    )

    if block_axes + 1 == len(shape):
        return right_chunks

    left_shape = shape[:-(block_axes+1)]
    left_chunks = np.array(np.meshgrid(*map(range, left_shape))).T.reshape(
        -1, len(left_shape)
    )

    return tuple(
        tuple(map(FullSlice.one, left, shape)) + tuple(right) for left in left_chunks for right in right_chunks
    )


def _len(obj: int | FullSlice) -> int:
    if isinstance(obj, FullSlice):
        return len(obj)

    return 1


def get_work_array(shape: tuple[int, ...],
                   slicer: tuple[FullSlice, ...],
                   dtype: np.dtype[_DT]) -> npt.NDArray[_DT]:
    if len(slicer) == 1 and slicer[0].is_whole_axis():
        return np.empty(shape, dtype=dtype)

    slicer_shape = tuple(_len(s) for s in slicer)
    return np.empty(slicer_shape, dtype=dtype)


def _read_array(arr: npt.NDArray[Any] | H5Array[Any],
                out: npt.NDArray[Any],
                source_sel: tuple[FullSlice, ...],
                dest_sel: tuple[FullSlice, ...]) -> None:
    if isinstance(arr, ch5mpy.H5Array):
        arr.read_direct(out, source_sel=source_sel, dest_sel=dest_sel)

    else:
        out[dest_sel] = arr[source_sel]


def iter_chunks_2(x1: npt.NDArray[Any] | H5Array[Any],
                  x2: npt.NDArray[Any] | H5Array[Any]) \
        -> Generator[tuple[tuple[FullSlice, ...], npt.NDArray[Any], npt.NDArray[Any] | Number], None, None]:
    # special case where x2 is a 0D array, iterate through chunks of x1 and always yield x2
    if x2.ndim == 0:
        max_mem_x1 = get_size(x1.MAX_MEM_USAGE) if isinstance(x1, ch5mpy.H5Array) else INF

        chunks = get_chunks(max_mem_x1, x1.shape, x1.dtype.itemsize)
        work_array_x1 = get_work_array(x1.shape, chunks[0], dtype=x1.dtype)

        for chunk in chunks:
            work_subset = tuple(c.shift_to_zero() for c in chunk)
            _read_array(x1, work_array_x1, chunk, work_subset)

            yield chunk, work_array_x1[work_subset], cast(Number, x2[()])

    # nD case
    else:
        if x1.shape != x2.shape:
            raise ValueError(f'Cannot iterate chunks of arrays with different shapes: {x1.shape} != {x2.shape}')

        max_mem_x1 = get_size(x1.MAX_MEM_USAGE) if isinstance(x1, ch5mpy.H5Array) else INF
        max_mem_x2 = get_size(x2.MAX_MEM_USAGE) if isinstance(x2, ch5mpy.H5Array) else INF

        chunks = get_chunks(min(max_mem_x1, max_mem_x2), x1.shape, max(x1.dtype.itemsize, x2.dtype.itemsize))
        work_array_x1 = get_work_array(x1.shape, chunks[0], dtype=x1.dtype)
        work_array_x2 = get_work_array(x1.shape, chunks[0], dtype=x2.dtype)

        for chunk in chunks:
            work_subset = tuple(c.shift_to_zero() for c in chunk)
            _read_array(x1, work_array_x1, chunk, work_subset)
            _read_array(x2, work_array_x2, chunk, work_subset)

            yield chunk, work_array_x1[work_subset], work_array_x2[work_subset]



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import numpy as np
from itertools import zip_longest

from numpy import typing as npt
from typing import Collection
from typing import Generator
from typing import Sequence
from typing import cast
from typing import Union
from typing import TypeVar
from typing import Iterable

from ch5mpy import Dataset
from ch5mpy._typing import SELECTOR
from ch5mpy.h5array.slice import FullSlice
from ch5mpy.utils import is_sequence

# ====================================================
# code
_DT = TypeVar("_DT", bound=np.generic)


def _get_array_sel(
    sel: tuple[Collection[int] | FullSlice, ...], index: int
) -> tuple[int | slice]:
    return cast(
        tuple[Union[int, slice]],
        tuple(slice(None) if isinstance(s, (Collection, FullSlice)) else index for s in sel)
    )


def _index_non_slice(elements: Iterable[Collection[int] | FullSlice]) -> int | None:
    for i, e in enumerate(elements):
        if not isinstance(e, FullSlice):
            return i

    return None


def _map_slice(
    elements: tuple[int | Sequence[int] | FullSlice | slice, ...]
) -> tuple[int | Sequence[int] | slice, ...]:
    return tuple(e.as_slice() if isinstance(e, FullSlice) else e for e in elements)


def _selection_iter(sel: tuple[Sequence[int] | FullSlice, ...]) -> Generator[
    tuple[tuple[int | Sequence[int] | slice, ...], tuple[int | slice, ...]], None, None
]:
    cut_index = _index_non_slice(sel)

    if len(sel) == 1 or cut_index is None:
        yield _map_slice(sel), _get_array_sel(sel, 0)

    else:
        pre = sel[:cut_index]
        post = sel[cut_index + 1:]
        cut = cast(Sequence[int], sel[cut_index])

        if len(cut) == 1:
            for p, arr_idx in _selection_iter(post):
                s = cut.start if isinstance(cut, FullSlice) else cut[0]

                yield _map_slice((*pre, s, *p)), (
                    *_get_array_sel(pre, 0),
                    *arr_idx,
                )

        else:
            for i, s in enumerate(cut):
                for p, arr_idx in _selection_iter(post):
                    yield _map_slice((*pre, s, *p)), (
                        *_get_array_sel(pre, 0),
                        i,
                        *arr_idx,
                    )


def read_from_dataset(dataset: Dataset[_DT],
                      selection: tuple[Sequence[int] | FullSlice, ...],
                      loading_array: npt.NDArray[_DT]) -> None:
    for dataset_idx, array_idx in _selection_iter(selection):
        dataset.read_direct(loading_array, source_sel=dataset_idx, dest_sel=array_idx)


def write_to_dataset(
        dataset: Dataset[_DT],
        values: npt.NDArray[_DT],
        selection: tuple[Sequence[int] | FullSlice, ...]
) -> None:
    if values.ndim == 0:
        values = values.reshape(1)

    for dataset_idx, array_idx in _selection_iter(selection):
        dataset.write_direct(values, source_sel=array_idx, dest_sel=dataset_idx)


def _check_bounds(max_: int, sel: Collection[int] | FullSlice, axis: int) -> None:
    valid = True

    if isinstance(sel, FullSlice):
        if sel.start < -max_ or sel.true_stop > max_:
            valid = False

    elif is_sequence(sel):
        if min(sel) < -max_ or max(sel) > max_:
            valid = False

    if not valid:
        raise IndexError(
            f"Index {sel} is out of bounds for axis {axis} with size {max_}."
        )


def _expand_index(index: tuple[SELECTOR, ...], shape: tuple[int, ...]) -> tuple[SELECTOR, ...]:
    exp_index: tuple[SELECTOR, ...] = tuple()
    i = 0

    for s in shape:
        if s == 1:
            exp_index += (0,)

        else:
            exp_index += (index[i],)
            i += 1
            if i == len(index):
                break

    return exp_index


def parse_selector(shape: tuple[int, ...],
                   index: SELECTOR | tuple[SELECTOR, ...]) \
        -> tuple[tuple[Sequence[int] | FullSlice, ...] | None, np.int_]:
    if index in (slice(None), ()):
        return None, np.product(shape)

    elif not isinstance(index, tuple):
        index = (index,)

    selection: tuple[Sequence[int] | FullSlice, ...] = ()

    for axis, (axis_len, axis_index) in enumerate(
        zip_longest(shape, _expand_index(index, shape), fillvalue=slice(None))
    ):
        if isinstance(axis_len, slice):
            raise RuntimeError

        parsed_axis_index: Union[FullSlice, Sequence[int]]

        if isinstance(axis_index, (slice, range)):
            parsed_axis_index = FullSlice(
                axis_index.start, axis_index.stop, axis_index.step, axis_len
            )

        elif is_sequence(axis_index):
            parsed_axis_index = axis_index

        elif isinstance(axis_index, int):
            parsed_axis_index = [axis_index]

        else:
            raise RuntimeError

        _check_bounds(axis_len, parsed_axis_index, axis)
        selection += (parsed_axis_index,)

    return selection, np.product([len(s) for s in selection])



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import numpy as np
from numpy import _NoValue as NoValue                                                       # type: ignore[attr-defined]
from functools import partial

import numpy.typing as npt
from typing import Any
from typing import Iterable
from typing import TYPE_CHECKING

from ch5mpy.h5array.functions.implement import implements
from ch5mpy.h5array.functions.apply import apply

if TYPE_CHECKING:
    from ch5mpy import H5Array


# ====================================================
# code
# ufuncs ----------------------------------------------------------------------
def _apply_ufunc(a: H5Array[Any],
                 out: tuple[H5Array[Any] | npt.NDArray[Any]] | None,
                 ufunc: partial[np.ufunc],
                 where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue,
                 dtype: npt.DTypeLike | None) -> Any:
    return apply(ufunc, '__set__', a,
                 out=None if out is None else out[0],
                 dtype=dtype,
                 initial=None,
                 where=where,
                 default_0D_output=False)


@implements(np.sin)
def sin(a: H5Array[Any],
        out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
        *,
        where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
        dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.sin, dtype=dtype), where, dtype)


@implements(np.cos)
def cos(a: H5Array[Any],
        out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
        *,
        where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
        dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.cos, dtype=dtype), where, dtype)


@implements(np.tan)
def tan(a: H5Array[Any],
        out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
        *,
        where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
        dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.tan, dtype=dtype), where, dtype)


@implements(np.arcsin)
def arcsin(a: H5Array[Any],
           out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
           *,
           where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
           dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.arcsin, dtype=dtype), where, dtype)


@implements(np.arccos)
def arccos(a: H5Array[Any],
           out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
           *,
           where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
           dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.arccos, dtype=dtype), where, dtype)


@implements(np.arctan)
def arctan(a: H5Array[Any],
           out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
           *,
           where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
           dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.arctan, dtype=dtype), where, dtype)


@implements(np.sinh)
def sinh(a: H5Array[Any],
         out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
         *,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
         dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.sinh, dtype=dtype), where, dtype)


@implements(np.cosh)
def cosh(a: H5Array[Any],
         out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
         *,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
         dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.cosh, dtype=dtype), where, dtype)


@implements(np.tanh)
def tanh(a: H5Array[Any],
         out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
         *,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
         dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.tanh, dtype=dtype), where, dtype)


@implements(np.arcsinh)
def arcsinh(a: H5Array[Any],
            out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
            *,
            where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
            dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.arcsinh, dtype=dtype), where, dtype)


@implements(np.arccosh)
def arccosh(a: H5Array[Any],
            out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
            *,
            where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
            dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.arccosh, dtype=dtype), where, dtype)


@implements(np.arctanh)
def arctanh(a: H5Array[Any],
            out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
            *,
            where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
            dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.arctanh, dtype=dtype), where, dtype)


@implements(np.floor)
def floor(a: H5Array[Any],
          out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
          *,
          where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
          dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.floor, dtype=dtype), where, dtype)


@implements(np.ceil)
def ceil(a: H5Array[Any],
         out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
         *,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
         dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.ceil, dtype=dtype), where, dtype)


@implements(np.trunc)
def trunc(a: H5Array[Any],
          out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
          *,
          where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
          dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.trunc, dtype=dtype), where, dtype)


@implements(np.exp)
def exp(a: H5Array[Any],
        out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
        *,
        where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
        dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.exp, dtype=dtype), where, dtype)


@implements(np.expm1)
def expm1(a: H5Array[Any],
          out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
          *,
          where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
          dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.expm1, dtype=dtype), where, dtype)


@implements(np.exp2)
def exp2(a: H5Array[Any],
         out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
         *,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
         dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.exp2, dtype=dtype), where, dtype)


@implements(np.log)
def log(a: H5Array[Any],
        out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
        *,
        where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
        dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.log, dtype=dtype), where, dtype)


@implements(np.log10)
def log10(a: H5Array[Any],
          out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
          *,
          where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
          dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.log10, dtype=dtype), where, dtype)


@implements(np.log2)
def log2(a: H5Array[Any],
         out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
         *,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
         dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.log2, dtype=dtype), where, dtype)


@implements(np.log1p)
def log1p(a: H5Array[Any],
          out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
          *,
          where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
          dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.log1p, dtype=dtype), where, dtype)


@implements(np.positive)
def positive(a: H5Array[Any],
             out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
             *,
             where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
             dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.positive, dtype=dtype), where, dtype)


@implements(np.negative)
def negative(a: H5Array[Any],
             out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
             *,
             where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
             dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.negative, dtype=dtype), where, dtype)


@implements(np.sqrt)
def sqrt(a: H5Array[Any],
         out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
         *,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
         dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.sqrt, dtype=dtype), where, dtype)


@implements(np.cbrt)
def cbrt(a: H5Array[Any],
         out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
         *,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
         dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.cbrt, dtype=dtype), where, dtype)


@implements(np.square)
def square(a: H5Array[Any],
           out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
           *,
           where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
           dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.square, dtype=dtype), where, dtype)


@implements(np.absolute)
def absolute(a: H5Array[Any],
             out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
             *,
             where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
             dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.absolute, dtype=dtype), where, dtype)


@implements(np.fabs)
def fabs(a: H5Array[Any],
         out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
         *,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
         dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.fabs, dtype=dtype), where, dtype)


@implements(np.sign)
def sign(a: H5Array[Any],
         out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
         *,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue,
         dtype: npt.DTypeLike | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.sign, dtype=dtype), where, dtype)


@implements(np.isfinite)
def isfinite(a: H5Array[Any],
             out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
             *,
             where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue) -> Any:
    return _apply_ufunc(a, out, partial(np.isfinite), where, dtype=bool)


@implements(np.isinf)
def isinf(a: H5Array[Any],
          out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
          *,
          where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue) -> Any:
    return _apply_ufunc(a, out, partial(np.isinf), where, dtype=bool)


@implements(np.isnan)
def isnan(a: H5Array[Any],
          out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None,
          *,
          where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue) -> Any:
    return _apply_ufunc(a, out, partial(np.isnan), where, dtype=bool)


@implements(np.isneginf)
def isneginf(a: H5Array[Any],
             out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.isneginf), where=NoValue, dtype=bool)


@implements(np.isposinf)
def isposinf(a: H5Array[Any],
             out: tuple[H5Array[Any] | npt.NDArray[Any]] | None = None) -> Any:
    return _apply_ufunc(a, out, partial(np.isposinf), where=NoValue, dtype=bool)


# numpy functions -------------------------------------------------------------
@implements(np.prod)
def prod(a: H5Array[Any],
         axis: int | Iterable[int] | tuple[int] | None = None,
         dtype: npt.DTypeLike | None = None,
         out: H5Array[Any] | npt.NDArray[Any] | None = None,
         keepdims: bool = False,
         initial: int | float | complex | None = None,
         where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue) -> Any:
    return apply(partial(np.prod, keepdims=keepdims, dtype=dtype, axis=axis), '__imul__', a, out,
                 dtype=dtype, initial=1, where=where)


@implements(np.sum)
def sum(a: H5Array[Any],
        axis: int | Iterable[int] | tuple[int] | None = None,
        dtype: npt.DTypeLike | None = None,
        out: H5Array[Any] | npt.NDArray[Any] | None = None,
        keepdims: bool = False,
        initial: int | float | complex | None = None,
        where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue) -> Any:
    initial = 0 if initial is None else initial

    return apply(partial(np.sum, keepdims=keepdims, dtype=dtype, axis=axis), '__iadd__', a, out,
                 dtype=dtype, initial=initial, where=where)


@implements(np.all)
def all(a: H5Array[Any],
        axis: int | Iterable[Any] | tuple[int] | None = None,
        out: H5Array[Any] | npt.NDArray[Any] | None = None,
        keepdims: bool = False,
        *,
        where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue) -> npt.NDArray[Any] | bool:
    return apply(partial(np.all, keepdims=keepdims, axis=axis), '__iand__', a, out,  # type: ignore[no-any-return]
                 dtype=bool, initial=True, where=where)


@implements(np.any)
def any(a: H5Array[Any],
        axis: int | Iterable[Any] | tuple[int] | None = None,
        out: H5Array[Any] | npt.NDArray[Any] | None = None,
        keepdims: bool = False,
        *,
        where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue = NoValue) -> npt.NDArray[Any] | bool:
    return apply(partial(np.any, keepdims=keepdims, axis=axis), '__ior__', a, out,  # type: ignore[no-any-return]
                 dtype=bool, initial=False, where=where)



# coding: utf-8

# ====================================================
# imports
from typing import Any
from typing import Sequence
from typing import Collection
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import TypeGuard


# ====================================================
# code
def is_sequence(obj: Any) -> 'TypeGuard[Sequence[Any]]':
    """Is the object a sequence of objects ? (excluding strings and byte objects.)"""
    return isinstance(obj, Collection) and hasattr(obj, '__getitem__') and not isinstance(
        obj, (str, bytes, bytearray, memoryview)
    )



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import h5py
import numpy as np
from abc import ABC
from abc import abstractmethod
from numbers import Number
from h5py._hl.base import ValuesViewHDF5
from h5py._hl.base import ItemsViewHDF5

import numpy.typing as npt
from typing import Any
from typing import cast
from typing import TypeVar
from typing import Generic

from ch5mpy._typing import SELECTOR
from ch5mpy.pickle.wrap import PickleableH5PyObject


# ====================================================
# code
_T = TypeVar('_T', bound=np.generic)


class _GroupManagerMixin(h5py.Group, ABC):
    """Mixin for File and Group objects which can access and create groups on H5 files."""

    @abstractmethod
    def _wrap(self, obj: Any) -> Any:
        """Wrap an object accessed in this group with our custom classes."""
        pass

    def __getitem__(self, name: str | bytes) -> Any:                                        # type: ignore[override]
        return self._wrap(h5py.Group.__getitem__(self, name))                               # type: ignore[index]

    def values(self) -> ValuesViewHDF5[Group | Dataset[Any]]:                               # type: ignore[override]
        return super().values()                                                             # type: ignore[return-value]

    def items(self) -> ItemsViewHDF5[str, Group | Dataset[Any]]:                            # type: ignore[override]
        return super().items()                                                              # type: ignore[return-value]

    def create_group(self, name: str, track_order: bool | None = None) -> Group:
        """
        Create and return a new subgroup.

        Args:
            name: may be absolute or relative. Fails if the target name already exists.
            track_order: Track dataset/group/attribute creation order under this group if True. If None use global
            default h5.get_config().track_order.
        """
        group = super().create_group(name, track_order=track_order)

        return cast(Group, self._wrap(group))


def _h5py_wrap_type(obj: Any) -> Any:
    """Produce our objects instead of h5py default objects"""
    if isinstance(obj, h5py.Dataset):
        return Dataset(obj.id)
    elif isinstance(obj, h5py.File):
        return File(obj.id)
    elif isinstance(obj, h5py.Group):
        return Group(obj.id)
    elif isinstance(obj, h5py.Datatype):
        return obj  # Not supported for pickling yet. Haven't really thought about it
    else:
        return obj  # Just return, since we want to wrap h5py.Group.get too


class Dataset(Generic[_T], PickleableH5PyObject, h5py.Dataset):
    """Mix in our pickling class"""

    def __getitem__(self,                                                                       # type: ignore[override]
                    arg: SELECTOR | tuple[SELECTOR, ...],
                    new_dtype: npt.DTypeLike | None = None) -> Number | str | npt.NDArray[_T]:
        return super().__getitem__(arg, new_dtype)

    def __setitem__(self, arg: SELECTOR | tuple[SELECTOR, ...], val: Any) -> None:
        super().__setitem__(arg, val)

    @property
    def dtype(self) -> np.dtype[_T]:
        return self.id.dtype                                                                # type: ignore[return-value]


class Group(PickleableH5PyObject, _GroupManagerMixin):
    """Overwrite group to allow pickling, and to create new groups and datasets
    of the right type (i.e. the ones defined in this module).
    """

    def _wrap(self, obj: Any) -> Any:
        obj = _h5py_wrap_type(obj)

        # If it is a group or dataset copy the current file info in
        if isinstance(obj, Group) or isinstance(obj, Dataset):
            obj.file_info = self.file_info

        return obj


class File(PickleableH5PyObject, _GroupManagerMixin, h5py.File):
    """A subclass of h5py.File that implements pickling.
    Pickling is done not with __{get,set}state__ but with __getnewargs_ex__
    which produces the arguments to supply to the __new__ method.
    """

    # noinspection PyMissingConstructor
    def __init__(self, *args: Any, **kwargs: Any):
        # Store args and kwargs for pickling
        self.init_args = args
        self.init_kwargs = kwargs

    def __new__(cls, *args: Any, **kwargs: Any) -> File:
        """Create a new File object with the h5 open function."""
        self = super().__new__(cls)
        h5py.File.__init__(self, *args, **kwargs)

        return self

    def _wrap(self, obj: Any) -> Any:
        obj = _h5py_wrap_type(obj)

        # If it is a group or dataset copy the current file info in
        if isinstance(obj, Group) or isinstance(obj, Dataset):
            obj.file_info = self

        return obj

    def __getstate__(self) -> None:
        pass

    def __getnewargs_ex__(self) -> tuple[tuple[Any, ...], dict[str, Any]]:
        kwargs = self.init_kwargs.copy()

        if len(self.init_args) > 1 and self.init_args[1] == "w":
            return (self.init_args[0], "r+", *self.init_args[2:]), kwargs

        return self.init_args, kwargs



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import numpy as np
import numpy.lib.mixins
from numbers import Number

import numpy.typing as npt
from typing import Any
from typing import cast
from typing import Generic
from typing import TypeVar
from typing import Iterator
from typing import Generator
from typing import TYPE_CHECKING
from ch5mpy._typing import NP_FUNC
from ch5mpy._typing import SELECTOR

from ch5mpy import Dataset
from ch5mpy.h5array import repr
from ch5mpy.h5array.inplace import get_chunks
from ch5mpy.h5array.inplace import get_work_array
from ch5mpy.h5array.functions import HANDLED_FUNCTIONS
from ch5mpy.h5array.io import parse_selector
from ch5mpy.h5array.io import write_to_dataset
from ch5mpy.h5array.slice import FullSlice
from ch5mpy.h5array.slice import map_slice

if TYPE_CHECKING:
    from ch5mpy.h5array.subset import H5ArrayView


# ====================================================
# code
_T = TypeVar("_T", bound=np.generic)


class H5Array(Generic[_T], numpy.lib.mixins.NDArrayOperatorsMixin):
    """Wrapper around Dataset objects to interface with numpy's API."""

    MAX_MEM_USAGE: int | str = "250M"

    # region magic methods
    def __init__(self, dset: Dataset[_T]):
        if not isinstance(dset, Dataset):
            raise TypeError(
                f"Object of type '{type(dset)}' is not supported by H5Array."
            )

        self._dset = dset

    def __repr__(self) -> str:
        return (
            f"H5Array({repr.print_dataset(self, end='', padding=8, padding_skip_first=True)}, "
            f"shape={self.shape}, dtype={self._dset.dtype})"
        )

    def __str__(self) -> str:
        return repr.print_dataset(self, sep="")

    def __getitem__(self, index: SELECTOR | tuple[SELECTOR, ...]) -> _T | H5Array[_T] | H5ArrayView[_T]:
        from ch5mpy.h5array.subset import H5ArrayView

        selection, nb_elements = parse_selector(self.shape, index)

        if nb_elements == 1:
            return cast(_T, self._dset[cast(SELECTOR, index)])

        elif selection is None:
            return H5Array(dset=self._dset)

        else:
            return H5ArrayView(dset=self._dset, sel=selection)

    def __setitem__(self, index: SELECTOR | tuple[SELECTOR, ...], value: Any) -> None:
        selection, nb_elements = parse_selector(self.shape, index)

        try:
            value_arr = np.array(value, dtype=self.dtype)

        except ValueError:
            raise ValueError(f'Could set value of type {type(value)} in H5Array of type {self.dtype}.')

        if nb_elements != value_arr.size:
            raise ValueError(f"{' x '.join(map(str, self.shape if selection is None else map(len, selection)))} "
                             f"values were selected but {' x '.join(map(str, value_arr.shape))} were given.")

        if selection is None:
            self._dset[()] = value_arr

        else:
            write_to_dataset(self._dset, value_arr, selection)

    def __len__(self) -> int:
        return len(self._dset)

    def __iter__(self) -> Iterator[_T | npt.NDArray[_T] | H5Array[_T] | H5ArrayView[_T]]:
        for i in range(self.shape[0]):
            yield self[i]

    def __contains__(self, item: Any) -> bool:
        raise NotImplementedError

    def _inplace_operation(self, func: NP_FUNC, value: Any) -> H5Array[_T]:
        if self.shape == ():
            self._dset[:] = func(self._dset[:], value)

        else:
            chunks = get_chunks(self.MAX_MEM_USAGE, self.shape, self.dtype.itemsize)
            work_array = get_work_array(self.shape, chunks[0], dtype=self.dtype)

            for chunk in chunks:
                work_subset = map_slice(c.shift_to_zero() for c in chunk)
                dataset_subset = map_slice(chunk)

                self._dset.read_direct(work_array, source_sel=dataset_subset, dest_sel=work_subset)
                func(work_array, value, out=work_array)
                self._dset.write_direct(work_array, source_sel=work_subset, dest_sel=dataset_subset)

        return self

    def __add__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return self._dset[()] + other                                                      # type: ignore[no-any-return]

    def __iadd__(self, other: Any) -> H5Array[_T]:
        return self._inplace_operation(np.add, other)

    def __sub__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return self._dset[()] - other                                                      # type: ignore[no-any-return]

    def __isub__(self, other: Any) -> H5Array[_T]:
        return self._inplace_operation(np.subtract, other)

    def __mul__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return self._dset[()] * other                                                      # type: ignore[no-any-return]

    def __imul__(self, other: Any) -> H5Array[_T]:
        return self._inplace_operation(np.multiply, other)

    def __truediv__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return self._dset[()] / other                                                      # type: ignore[no-any-return]

    def __itruediv__(self, other: Any) -> H5Array[_T]:
        return self._inplace_operation(np.divide, other)

    def __mod__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return self._dset[()] % other                                                      # type: ignore[no-any-return]

    def __imod__(self, other: Any) -> H5Array[_T]:
        return self._inplace_operation(np.mod, other)

    def __pow__(self, other: Any) -> Number | str | npt.NDArray[Any]:
        return self._dset[()] ** other                                                     # type: ignore[no-any-return]

    def __ipow__(self, other: Any) -> H5Array[_T]:
        return self._inplace_operation(np.power, other)

    def __or__(self, other: Any) -> Number | npt.NDArray[Any]:
        return self._dset[()] | other                                                      # type: ignore[no-any-return]

    def __ior__(self, other: Any) -> H5Array[_T]:
        return self._inplace_operation(np.logical_or, other)

    def __and__(self, other: Any) -> Number | npt.NDArray[Any]:
        return self._dset[()] & other                                                      # type: ignore[no-any-return]

    def __iand__(self, other: Any) -> H5Array[_T]:
        return self._inplace_operation(np.logical_and, other)

    def __invert__(self) -> Number | npt.NDArray[Any]:
        return ~self._dset[()]                                                                  # type: ignore[operator]

    def __xor__(self, other: Any) -> Number | npt.NDArray[Any]:
        return self._dset[()] ^ other                                                      # type: ignore[no-any-return]

    def __ixor__(self, other: Any) -> H5Array[_T]:
        return self._inplace_operation(np.logical_xor, other)

    # endregion

    # region interface
    def __array__(self, dtype: npt.DTypeLike | None = None) -> npt.NDArray[Any]:
        return np.array(self._dset).astype(dtype)

    def __array_ufunc__(self, ufunc: NP_FUNC, method: str, *inputs: Any, **kwargs: Any) \
            -> Any:
        if method == "__call__":
            if ufunc not in HANDLED_FUNCTIONS:
                return NotImplemented

            return HANDLED_FUNCTIONS[ufunc](*inputs, **kwargs)

        else:
            raise NotImplemented

    def __array_function__(
        self,
        func: NP_FUNC,
        types: tuple[type, ...],
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        if func not in HANDLED_FUNCTIONS:
            return NotImplemented

        return HANDLED_FUNCTIONS[func](*args, **kwargs)

    # endregion

    # region attributes
    @property
    def dset(self) -> Dataset[_T]:
        return self._dset

    @property
    def shape(self) -> tuple[int, ...]:
        return self._dset.shape

    @property
    def dtype(self) -> np.dtype[_T]:
        return self._dset.dtype

    @property
    def ndim(self) -> int:
        return len(self.shape)

    # endregion

    # region methods
    def astype(self, dtype: npt.DTypeLike) -> npt.NDArray[Any]:
        return np.array(self, dtype=dtype)

    def iter_chunks(self, keepdims: bool = False) \
            -> Generator[tuple[tuple[FullSlice, ...], npt.NDArray[_T]], None, None]:
        chunks = get_chunks(self.MAX_MEM_USAGE, self.shape, self.dtype.itemsize)
        work_array = get_work_array(self.shape, chunks[0], dtype=self.dtype)

        for chunk in chunks:
            work_subset = map_slice(c.shift_to_zero() for c in chunk)
            self._dset.read_direct(work_array, source_sel=map_slice(chunk), dest_sel=work_subset)

            if keepdims:
                res = work_array[work_subset]
                yield chunk, res.reshape((1,) * (self.ndim - res.ndim) + res.shape)

            else:
                yield chunk, work_array[work_subset]

    def read_direct(self,
                    dest: npt.NDArray[_T],
                    source_sel: tuple[FullSlice, ...],
                    dest_sel: tuple[FullSlice, ...]) -> None:
        self._dset.read_direct(dest, source_sel=map_slice(source_sel), dest_sel=map_slice(dest_sel))

    # endregion



# coding: utf-8
# Created on 13/12/2022 14:41
# Author : matteo

# ====================================================
# imports
from .objects import File, Group, Dataset
from .h5dict import H5Dict
from .h5array.h5array import H5Array
from .write import (
    write_attribute,
    write_attributes,
    write_dataset,
    write_datasets,
    write_object,
    write_objects,
)
from .names import H5Mode

# ====================================================
# code
__all__ = [
    "File",
    "Group",
    "Dataset",
    "H5Dict",
    "H5Array",
    "write_attribute",
    "write_attributes",
    "write_dataset",
    "write_datasets",
    "write_object",
    "write_objects",
    "H5Mode",
]



# coding: utf-8

# ====================================================
# imports
from ch5mpy.h5array.slice import FullSlice


# ====================================================
# code
def test_true_stop_step1():
    f = FullSlice(0, 3, 1, 10)
    assert f.true_stop == 2


def test_true_stop_step3():
    f = FullSlice(5, 13, 3, 20)
    assert f.true_stop == 11


def test_true_stop_no_len():
    f = FullSlice(0, 0, 1, 0)
    assert f.true_stop == 0


def test_length():
    f = FullSlice(0, 3, 1, 10)
    assert len(f) == 3



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import numpy as np

import numpy.typing as npt
from typing import Any
from typing import Iterable


# ====================================================
# code
class FullSlice:

    # region magic methods
    def __init__(
        self,
        start: int | None,
        stop: int | None,
        step: int | None,
        max_: int,
    ):
        self._start = start or 0
        self._step = step or 1
        self._stop = stop or max_
        self._max = max_

        if self._start < 0:
            self._start = self._stop + self._start

    def __repr__(self) -> str:
        if self.is_one_element():
            return f"{type(self).__name__}(<{self.start}> | {self._max})"

        if self.is_whole_axis():
            return f"{type(self).__name__}(* | {self._max})"

        return f"{type(self).__name__}({self._start}, {self._stop}, {self._step} | {self._max})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, FullSlice):
            raise NotImplementedError

        if self._start == other.start and self._stop == other.stop and self._step == other.step \
                and self._max == other.max:
            return True

        return False

    def __len__(self) -> int:
        return (self.true_stop - self._start) // self._step + 1

    @classmethod
    def whole_axis(cls, max_: int) -> FullSlice:
        return FullSlice(0, max_, 1, max_)

    @classmethod
    def one(cls,
            element: int,
            max_: int | None = None) -> FullSlice:
        max_ = element + 1 if max_ is None else max_
        return FullSlice(element, element+1, 1, max_)

    def __array__(self, dtype: npt.DTypeLike | None = None) -> npt.NDArray[Any]:
        return np.array(range(self.start, self._stop, self._step), dtype=dtype)

    # endregion

    # region attributes
    @property
    def start(self) -> int:
        return self._start

    @property
    def stop(self) -> int:
        return self._stop

    @property
    def true_stop(self) -> int:
        """Get the true last int in ths slice, if converted to a list."""
        if self._start == self._stop:
            return self._stop

        last = np.arange(self._stop - self._step, self._stop, dtype=int)
        return int(last[last % self._step == self._start % self._step][-1])

    @property
    def step(self) -> int:
        return self._step

    @property
    def max(self) -> int:
        return self._max

    # endregion

    # region predicates
    def is_whole_axis(self, max_: int | None = None) -> bool:
        max_ = max_ or self._max
        return self._start == 0 and self._step == 1 and self._stop == max_

    def is_one_element(self) -> bool:
        return len(self) == 1

    # endregion

    # region methods
    def as_slice(self) -> slice:
        return slice(self.start, self.stop, self.step)

    def shift_to_zero(self) -> FullSlice:
        return FullSlice(0, self.stop - self.start, self.step, self._max)

    # endregion


def map_slice(index: Iterable[FullSlice]) -> tuple[slice, ...]:
    return tuple(fs.as_slice() for fs in index)



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import importlib
import numpy as np

from typing import Callable

from ch5mpy._typing import NP_FUNC
from ch5mpy._typing import H5_FUNC


# ====================================================
# code
HANDLED_FUNCTIONS: dict[NP_FUNC | np.ufunc, H5_FUNC] = {}


def implements(np_function: NP_FUNC | np.ufunc) -> Callable[[H5_FUNC], H5_FUNC]:
    """Register an __array_function__ implementation for DiagonalArray objects."""

    def decorator(func: H5_FUNC) -> H5_FUNC:
        HANDLED_FUNCTIONS[np_function] = func
        return func

    return decorator


# manually import function implementations otherwise they are never imported
importlib.__import__("ch5mpy.h5array.functions.two_arrays")
importlib.__import__("ch5mpy.h5array.functions.element_wise")



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

from typing import Collection

import numpy as np

from typing import Any
from typing import Union
from typing import Callable

# ====================================================
# code
SELECTOR = Union[int, bool, slice, range, Collection[int], Collection[bool], Collection[np.bool_], tuple[()]]

NP_FUNC = Callable[..., Any]
H5_FUNC = Callable[..., Any]



# coding: utf-8

# ====================================================
# imports
import numpy as np


# ====================================================
# code
def test_should_get_shape(array):
    assert array.shape == (10, 10)


def test_should_get_dtype(array):
    assert array.dtype == np.float64


def test_should_print_repr(array):
    assert (
        repr(array) == "H5Array([[0.0, 1.0, 2.0, ..., 7.0, 8.0, 9.0],\n"
        "         [10.0, 11.0, 12.0, ..., 17.0, 18.0, 19.0],\n"
        "         [20.0, 21.0, 22.0, ..., 27.0, 28.0, 29.0],\n"
        "         ...,\n"
        "         [70.0, 71.0, 72.0, ..., 77.0, 78.0, 79.0],\n"
        "         [80.0, 81.0, 82.0, ..., 87.0, 88.0, 89.0],\n"
        "         [90.0, 91.0, 92.0, ..., 97.0, 98.0, 99.0]], shape=(10, 10), dtype=float64)"
    )


def test_should_convert_to_numpy_array(array):
    assert isinstance(np.asarray(array), np.ndarray)
    assert np.array_equal(np.asarray(array), np.arange(100).reshape((10, 10)))


def test_should_pass_numpy_ufunc(array):
    arr_2 = np.multiply(array, 2)
    assert np.array_equal(arr_2, np.arange(100).reshape((10, 10)) * 2)


def test_should_work_with_magic_operations(array):
    arr_2 = array + 2
    assert np.array_equal(arr_2, np.arange(100).reshape((10, 10)) + 2)


def test_should_sum_all_values(array):
    s = np.sum(array)
    assert s == 4950


def test_should_sum_along_axis(array):
    s = np.sum(array, axis=0)
    assert np.array_equal(
        s, np.array([450, 460, 470, 480, 490, 500, 510, 520, 530, 540])
    )


def test_should_add_inplace(array):
    array += 1
    assert np.array_equal(array, np.arange(100).reshape((10, 10)) + 1)


# def test_large_array(large_array):
#     large_array += 2


def test_should_get_single_element(array):
    assert array[1, 2] == 12


def test_should_get_whole_dset(array):
    assert np.array_equal(array[:], array)
    assert np.array_equal(array[()], array)


def test_should_print_view_repr(array):
    sub_arr = array[2:4, [0, 2, 3]]
    assert str(sub_arr) == "[[20.0 22.0 23.0],\n" " [30.0 32.0 33.0]]\n"


def test_should_get_view(array):
    sub_arr = array[2:4, [0, 2, 3]]
    assert np.array_equal(sub_arr, np.array([[20, 22, 23], [30, 32, 33]]))


def test_should_get_view_from_view(array):
    sub_arr = array[2:4, [0, 2, 3]]
    sub_sub_arr = sub_arr[1, [1, 2]]
    assert np.array_equal(sub_sub_arr, np.array([32, 33]))


def test_should_get_single_value_from_view(array):
    assert array[2:4, [0, 2, 3]][0, 1] == 22


def test_should_set_value_in_array(array):
    array[5, 7] = -1
    assert array[5, 7] == -1


def test_should_set_value_in_view(array):
    sub_arr = array[2:4, [0, 2, 3]]
    sub_arr[1, [1, 2]] = [-2, -3]
    assert np.array_equal(array[3, [2, 3]], [-2, -3])


def test_apply_all_function(array):
    assert not np.all(array)

    array += 1
    assert np.all(array)


def test_apply_any_function(array):
    assert np.any(array)



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import numpy as np
from numpy import _NoValue as NoValue                                                       # type: ignore[attr-defined]
from numbers import Number
from functools import partial

import numpy.typing as npt
from typing import Any
from typing import Union
from typing import Iterable
from typing import TYPE_CHECKING

import ch5mpy
from ch5mpy.h5array.inplace import iter_chunks_2
from ch5mpy.h5array.slice import map_slice
from ch5mpy.h5array.slice import FullSlice

if TYPE_CHECKING:
    from ch5mpy import H5Array


# ====================================================
# code
WHERE_SELECTION = Union[npt.NDArray[np.bool_], tuple[()]]


class Where:
    def __init__(self,
                 where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue,
                 shape: tuple[int, ...]):
        # TODO : make memory efficient (avoid broadcast and compute on the fly)
        self._where = None if where in (True, NoValue) else np.broadcast_to(where, shape)       # type: ignore[arg-type]

    def __getitem__(self, item: tuple[Any, ...] | slice) -> WHERE_SELECTION:
        if self._where is None:
            return ()

        return self._where[item]


def _get_output_array(out: H5Array[Any] | npt.NDArray[Any] | None,
                      shape: tuple[int, ...],
                      axis: tuple[int, ...],
                      keepdims: bool,
                      dtype: npt.DTypeLike | None,
                      initial: int | float | complex | None,
                      default: Any) -> H5Array[Any] | npt.NDArray[Any]:
    if keepdims:
        expected_shape = tuple(s if i not in axis else 1 for i, s in enumerate(shape))

    else:
        expected_shape = tuple(s for i, s in enumerate(shape) if i not in axis)

    if out is None:
        out = np.empty(shape=expected_shape, dtype=dtype)

        if default is not None:
            out[()] = default

    else:
        ndim = len(expected_shape)

        if out.ndim != ndim:
            raise ValueError(f'Output array has the wrong number of dimensions: Found {out.ndim} but expected {ndim}')

        if out.shape != expected_shape:
            raise ValueError(f'Output array has the wrong shape: Found {out.shape} but expected {expected_shape}')

    if initial is not None:
        out[()] = initial

    return out


def _as_tuple(axis: int | Iterable[int] | tuple[int, ...] | None,
              ndim: int,
              default_0D_output: bool) -> tuple[int, ...]:
    if axis is None:
        return tuple(range(ndim)) if default_0D_output else ()

    elif not isinstance(axis, Iterable):
        return axis,

    return tuple(axis)


def _get_indices(index: tuple[FullSlice, ...],
                 axis: tuple[int, ...],
                 where_compute: Where,
                 where_output: Where,
                 output_ndim: int) -> tuple[WHERE_SELECTION, tuple[slice, ...] | tuple[()], WHERE_SELECTION]:
    # compute on whole array at once
    if len(index) == 1 and index[0].is_whole_axis():
        return where_compute[:], (), where_output[index]

    where_to_compute = where_compute[map_slice(index)]

    # 0D output array (no chunk selection, no where selection)
    if output_ndim == 0:
        return where_to_compute, (), ()

    # nD output array
    selected_index = map_slice(tuple(e for i, e in enumerate(index) if i not in axis))
    return where_to_compute, selected_index, where_output[selected_index]


def _apply_operation(operation: str,
                     dest: H5Array[Any] | npt.NDArray[Any],
                     chunk_selection: tuple[slice, ...] | tuple[()],
                     where_to_output: WHERE_SELECTION,
                     values: npt.NDArray[Any]) -> None:
    # Here I have to explicitly write out each operation using +=, *=, ... operators instead of using
    # getattr(..., operation)(...) because dest does not get modified in that way

    if operation == '__set__':
        dest[chunk_selection][where_to_output] = values[where_to_output]

    elif operation == '__iadd__':
        if dest.ndim == 0:
            dest += values                                                                    # type: ignore[assignment]

        else:
            dest[chunk_selection][where_to_output] += values[where_to_output]

    elif operation == '__imul__':
        if dest.ndim == 0:
            dest *= values                                                                    # type: ignore[assignment]

        else:
            dest[chunk_selection][where_to_output] *= values[where_to_output]

    elif operation == '__iand__':
        if dest.ndim == 0:
            dest &= values                                                                    # type: ignore[assignment]

        else:
            dest[chunk_selection][where_to_output] &= values[where_to_output]

    elif operation == '__ior__':
        if dest.ndim == 0:
            dest |= values                                                                    # type: ignore[assignment]

        else:
            dest[chunk_selection][where_to_output] |= values[where_to_output]

    else:
        raise NotImplementedError(f"Do not know how to apply operation '{operation}'")


def apply(func: partial[np.ufunc],
          operation: str,
          a: H5Array[Any],
          out: H5Array[Any] | npt.NDArray[Any] | None,
          *,
          dtype: npt.DTypeLike | None,
          initial: int | float | complex | None,
          where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue,
          default_0D_output: bool = True) -> Any:
    axis = _as_tuple(func.keywords.get('axis', None), a.ndim, default_0D_output)
    output_array = _get_output_array(out, a.shape, axis, func.keywords.get('keepdims', False), dtype, initial, None)

    if where is not False:
        where_compute = Where(where, a.shape)
        where_output = Where(where, output_array.shape)

        for index, chunk in a.iter_chunks(keepdims=True):
            where_to_compute, chunk_selection, where_to_output = \
                _get_indices(index, axis, where_compute, where_output, output_array.ndim)
            result = np.array(func(chunk, where=True if where_to_compute == () else where_to_compute),
                              dtype=output_array.dtype)
            _apply_operation(operation, output_array, chunk_selection, where_to_output, result)

    if out is None and output_array.ndim == 0:
        return output_array[()]

    return output_array


def apply_2(func: np.ufunc,
            a: H5Array[Any],
            b: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
            *,
            out: H5Array[Any] | npt.NDArray[Any] | None,
            default: Any,
            dtype: npt.DTypeLike | None,
            where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue) -> Any:
    output_array = _get_output_array(out, a.shape, (), False, dtype, None, default)

    if where is not False:
        where_compute = Where(where, a.shape)
        where_output = Where(where, output_array.shape)

        if not isinstance(b, (np.ndarray, ch5mpy.H5Array)):
            b = np.array(b)

        for index, chunk_x1, chunk_x2 in iter_chunks_2(a, b):
            where_to_compute, chunk_selection, where_to_output = \
                _get_indices(index, (), where_compute, where_output, output_array.ndim)
            result = np.array(func(chunk_x1, chunk_x2, where=True if where_to_compute == () else where_to_compute),
                              dtype=output_array.dtype)
            _apply_operation('__set__', output_array, chunk_selection, where_to_output, result)

    if out is None and output_array.ndim == 0:
        return output_array[()]

    return output_array



# coding: utf-8

# ====================================================
# imports
from __future__ import annotations

import numpy as np

import numpy.typing as npt
from typing import Any
from typing import cast
from typing import TypeVar
from typing import Sequence
from ch5mpy._typing import SELECTOR
from ch5mpy._typing import NP_FUNC

import ch5mpy.h5array.h5array as h5array
from ch5mpy import Dataset
from ch5mpy.h5array.io import parse_selector
from ch5mpy.h5array.io import read_from_dataset
from ch5mpy.h5array.io import write_to_dataset
from ch5mpy.h5array.slice import FullSlice
from ch5mpy.h5array.slice import map_slice

# ====================================================
# code
_T = TypeVar("_T", bound=np.generic)
_DT = TypeVar("_DT", bound=np.generic)


def _cast_selection(
        selection: tuple[Sequence[int] | FullSlice, ...],
        on: tuple[Sequence[int] | FullSlice, ...],
) -> tuple[Sequence[int] | FullSlice, ...]:
    casted_indices: tuple[Sequence[int] | FullSlice, ...] = ()

    for s, o in zip(selection, on):
        if isinstance(s, FullSlice) and \
                s.is_whole_axis(o.max if isinstance(o, FullSlice) else s.max):
            casted_indices += (o,)

        elif isinstance(o, FullSlice) and o.is_whole_axis():
            casted_indices += (s,)

        else:
            casted_indices += (np.array(o)[np.array(s)],)

    return casted_indices


class H5ArrayView(h5array.H5Array[_T]):
    """A view on a H5Array."""

    # region magic methods
    def __init__(self, dset: Dataset[_T], sel: tuple[Sequence[int] | FullSlice, ...]):
        super().__init__(dset)
        self._selection = sel

    def __getitem__(self, index: SELECTOR | tuple[SELECTOR, ...]) -> _T | H5ArrayView[_T]:
        selection, nb_elements = parse_selector(self.shape_selection, index)

        if selection is None:
            return H5ArrayView(dset=self._dset, sel=self._selection)

        elif nb_elements == 1:
            loading_array = np.empty(1, dtype=self.dtype)
            read_from_dataset(self._dset, _cast_selection(selection, on=self._selection), loading_array)

            return cast(_T, loading_array[0])

        return H5ArrayView(dset=self._dset, sel=_cast_selection(selection, on=self._selection))

    def __setitem__(self, index: SELECTOR | tuple[SELECTOR, ...], value: Any) -> None:
        selection, nb_elements = parse_selector(self.shape_selection, index)

        try:
            value_arr = np.array(value, dtype=self.dtype)

        except ValueError:
            raise ValueError(f'Could set value of type {type(value)} in H5Array of type {self.dtype}.')

        if nb_elements != value_arr.size:
            raise ValueError(f"{' x '.join(map(str, self.shape if selection is None else map(len, selection)))} "
                             f"values were selected but {' x '.join(map(str, value_arr.shape))} were given.")

        if selection is None:
            selection_as_slices: tuple[Sequence[int] | slice, ...] = \
                tuple(e.as_slice() if isinstance(e, FullSlice) else e for e in self._selection)
            self._dset[selection_as_slices] = value_arr

        else:
            write_to_dataset(self._dset, value_arr,  _cast_selection(selection, on=self._selection))

    def __len__(self) -> int:
        return self.shape[0]

    def __contains__(self, item: Any) -> bool:
        raise NotImplementedError

    def _inplace_operation(self, func: NP_FUNC, value: Any) -> H5ArrayView[_T]:
        raise NotImplementedError

    # endregion

    # region interface
    def __array__(self, dtype: npt.DTypeLike | None = None) -> npt.NDArray[Any]:
        if dtype is None:
            dtype = self.dtype

        loading_array = np.empty(self.shape, dtype)
        read_from_dataset(self._dset, self._selection, loading_array)

        return loading_array

    # endregion

    # region attributes
    @property
    def shape_selection(self) -> tuple[int, ...]:
        return tuple(len(axis_sel) for axis_sel in self._selection)

    @property
    def shape(self) -> tuple[int, ...]:
        return tuple(len(axis_sel) for axis_sel in self._selection if len(axis_sel) > 1)

    # endregion

    # region methods
    def read_direct(self,
                    dest: npt.NDArray[_T],
                    source_sel: tuple[FullSlice, ...],
                    dest_sel: tuple[FullSlice, ...]) -> None:
        nb_whole_axes_before = (self._dset.ndim - len(source_sel))
        whole_axes_before = tuple(FullSlice.whole_axis(s) for s in self.shape[:nb_whole_axes_before])
        nb_whole_axes_after = (len(whole_axes_before) + len(source_sel))
        whole_axes_after = tuple(FullSlice.whole_axis(s) for s in self.shape[nb_whole_axes_after:])

        source_sel_casted = _cast_selection(whole_axes_before + source_sel + whole_axes_after, on=self._selection)
        read_from_dataset(self._dset, source_sel_casted, dest[map_slice(dest_sel)])

    # endregion
